# Copyright 2023 - Compute Heavy Industries Incorporated
# This work is released, distributed, and licensed under AGPLv3.

import abc
import json
import uuid
import typing
import jsonschema
import multiprocessing

from gonk.core import exceptions
from gonk.core import events

### Machine ###
class Machine:
    """Machine coordinates event validation and consumption."""
    def __init__(self):
        self.validators: list[Validator] = []
        """:class:`Validator`\s that have been registered."""
        self.consumers: list[Consumer] = []
        """:class:`Consumer`\s that have been registered."""
        self.lock = multiprocessing.Lock()
        """**DO NOT CONSIDER THIS THREAD SAFE WITHOUT FURTHER TESTING**"""

    def process_event(self, event):
        """Runs registered validators and consumers."""
        with self.lock:
            for validator in self.validators:
                validator.validate(event)

            for consumer in self.consumers:
                consumer.consume(event)

    def register(self, worker):
        """Registers a class instance as a validator, consumer, or both."""
        registered = False

        if isinstance(worker, Validator):
            self.validators.append(worker)
            registered = True


        if isinstance(worker, Consumer):
            self.consumers.append(worker)
            registered = True

        if not registered:
            raise ValueError("not a consumer or validator")

class Validator(abc.ABC):
    """Abstract class for validators."""
    @abc.abstractmethod
    def validate(self, event):
        """Validate the event.

        Raises:
            ValidationError: Invalid event."""
        raise NotImplementedError("unimplemented method")

class Consumer(abc.ABC):
    """Abstract class for consumers."""
    @abc.abstractmethod
    def consume(self, event):
        """Consume the event."""
        raise NotImplementedError("unimplemented method")

### Record Keeper ###
class RecordKeeper(Consumer, Validator, abc.ABC):
    """A RecordKeeper acts as the :class:`gonk.core.events.Event` 
        storage for the dataset."""
    def validate(self, event: events.EventT):
        if self.exists(event.uuid):
            raise exceptions.ValidationError("event UUID already exists")

    def consume(self, event: events.EventT):
        self.add(event)

    @abc.abstractmethod
    def add(self, event: events.EventT):
        """Add a :class:`gonk.core.events.Event` to storage."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def exists(self, uuid_: uuid.UUID) -> bool:
        """Check :class:`gonk.core.events.Event` existence."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def read(self, uuid_: uuid.UUID) -> events.Event:
        """Read a :class:`gonk.core.events.Event` from storage.

        Raises:
            ValueError: Event does not exist."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def next(self, uuid_: uuid.UUID | None) -> uuid.UUID | None:
        """Get the next :class:`gonk.core.events.Event`'s 
            UUID in the event history.

        Returns:
            The next event UUID in the event history 
                or None if the provided UUID is the tail.

        Raises:
            ValueError: Event does not exist."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def tail(self) -> uuid.UUID | None:
        """Get the last :class:`gonk.core.events.Event`'s 
            UUID in the event history.

        Returns:
            The last event UUID in the event history 
                or None when there are no events."""
        raise NotImplementedError("unimplemented method")

### Depot ###
class Depot(abc.ABC):
    """A Depot acts as the object (and annotation) storage for a dataset.

        The Depot interface is a little clunky. It was designed to be 
        compatible with S3's multipart upload. This was done to enable 
        working with very large objects as well as bittorrent-style 
        block-based file sharing."""
    @abc.abstractmethod
    def exists(self, identifier: events.Identifier):
        """Check whether an object exists."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def reserve(self, identifier: events.Identifier, size: int):
        """Reserve an object identifier for writing.

        Raises:
            StorageError: Identifier already exists."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def write(self, identifier: events.Identifier, offset: int, buf: bytes):
        """Write bytes to an object at an `offset`.

        Raises:
            StorageError: Identifier not found.
            StorageError: Object finalized."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def finalize(self, identifier: events.Identifier):
        """Finalize an object and make it readable.

        Raises:
            StorageError: Identifier not found.
            StorageError: Object already finalized."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def read(self, identifier: events.Identifier, offset: int, size: int):
        """Read `size` bytes of an object from `offset`.

        Raises:
            StorageError: Identifier not found.
            StorageError: Object not yet finalized."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def purge(self, identifier: events.Identifier):
        """Delete the object.

        Raises:
            StorageError: Identifier not found."""
        raise NotImplementedError("unimplemented method")

### State ###
class SchemaInfo:
    """Informational class for schemas."""
    def __init__(self, name: str, uuid_: uuid.UUID, versions: int):
        self.name: str = name
        """Schema name."""
        self.uuid: uuid.UUID = uuid_
        """Schema UUID."""
        self.versions: int = versions
        """Number of versions."""

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        return {
            "name": self.name,
            "uuid": str(self.uuid),
            "versions": self.versions,
        }

class ObjectInfo:
    """Informational class for :class:`gonk.core.events.Object`\s."""
    def __init__(self, uuid_: uuid.UUID, versions: int):
        self.uuid: uuid.UUID = uuid_
        """Object UUID."""
        self.versions: int = versions
        """Number of versions."""

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        return {
            "uuid": str(self.uuid),
            "versions": self.versions,
        }

class AnnotationInfo:
    """Informational class for :class:`gonk.core.events.Annotation`\s."""
    def __init__(self, uuid_: uuid.UUID, versions: int):
        self.uuid: uuid.UUID = uuid_
        """Annotation UUID."""
        self.versions: int = versions
        """Number of versions."""

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        return {
            "uuid": str(self.uuid),
            "versions": self.versions,
        }

class EventInfo:
    """Informational class for :class:`gonk.core.events.Event`\s."""
    def __init__(self, uuid_: uuid.UUID, type_: str, review: None|str):
        self.uuid: uuid.UUID = uuid_
        """Event UUID."""
        self.type: str = type_
        """Event type (class name)."""
        self.review: None|str = review
        """The review status of the event.

        For unreviewable events, :class:`gonk.core.events.OwnerEvent` 
        and :class:`gonk.core.events.ReviewEvent`, this field is None.
        For reviewable events, :class:`gonk.core.events.ObjectEvent` and 
        :class:`gonk.core.events.AnnotationEvent`, this field can be
        ``PENDING``, ``ACCEPTED``, or ``REJECTED``."""

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        return {
            "uuid": str(self.uuid),
            "type": self.type,
            "review": self.review,
        }

    def __repr__(self):
        return f"EventInfo({self.uuid}, {self.type}, {self.review})"

class NamedIdentifier(events.Identifier):
    """Return type for schemas and objects."""
    def __init__(self, uuid_: uuid.UUID, version: int, name: str):
        super().__init__(uuid_, version)
        self.name = name

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        return {
            "uuid": str(self.uuid),
            "version": self.version,
            "name": self.name,
        }

class State(Validator, Consumer, abc.ABC):
    """The state interface tracks the current state of the dataset.

    The information it stores in implementation dependent but must enable 
    event validation as well as provide the API that the web service 
    requires to run."""
    @abc.abstractmethod
    def events_by_object(self, 
        identifier: events.Identifier) -> list[EventInfo]:
        """Get the events for a given object."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def events_by_annotation(self, 
        identifier: events.Identifier) -> list[EventInfo]:
        """Get the events for a given annotation."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def events_all(self, after: None|uuid.UUID = None) -> list[EventInfo]:
        """Get a page of 25 events.

        Args:
            after: Event UUID to page after or None for the first page."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def annotations_all(self, uuid_: None|uuid.UUID = None, 
        after: None|uuid.UUID = None) -> list[AnnotationInfo]:
        """Get a page of 25 annotations.

        Only one of ``uuid_`` and ``after`` may be provided.

        Args:
            uuid_: The annotation UUID to get information about.
            after: Annotation UUID to page after or None for the first page."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def annotations_by_object(self, 
        object_identifier: events.Identifier) -> list[AnnotationInfo]:
        """Gets the annotations for a given object."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def annotations_by_status(self, status: str, 
        after: None|uuid.UUID = None) -> list[events.Identifier]:
        """Gets the annotations for a given status.

        Args:
            status: ``pending``, ``accepted``, ``rejected``, or ``deleted``.
            after: Annotation UUID to page after or None for the first page."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def annotation(self, 
        identifier: events.Identifier) -> events.Annotation | None:
        """Get an Annotation for a given identifier."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def objects_all(self, uuid_: None|uuid.UUID = None, 
        after: None|uuid.UUID = None) -> list[ObjectInfo]:
        """Get a page of 25 objects.

        Only one of ``uuid_`` and ``after`` may be provided.

        Args:
            uuid_: The object UUID to get information about.
            after: Object UUID to page after or None for the first page."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def objects_by_annotation(self, 
        annotation_uuid: uuid.UUID) -> list[events.Identifier]:
        """Get the objects for a given annotation."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def objects_by_status(self, status: str, 
        after: None|uuid.UUID = None) -> list[events.Identifier]:
        """Gets the objects for a given status.

        Args:
            status: ``pending``, ``accepted``, ``rejected``, or ``deleted``.
            after: Object UUID to page after or None for the first page."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def object(self, identifier: events.Identifier) -> events.Object | None:
        """Get an Object for a given identifier."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def schemas_all(self, name: None|str = None) -> list[SchemaInfo]:
        """Get all schemas.

        Args:
            name: If provided only information 
                about that schema will be returned."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def schemas_by_status(self, 
        status: str, after: None|uuid.UUID = None) -> list[NamedIdentifier]:
        """Gets the schemas for a given status.

        Args:
            status: ``pending``, ``accepted``, ``rejected``, or ``deprecated``.
            after: Schema UUID to page after or None for the first page."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def schema(self, name: str, version: int) -> events.Object | None:
        """Get a schema for a given name and version."""
        raise NotImplementedError("unimplemented method")
    
    @abc.abstractmethod
    def owners(self) -> list[str]:
        """Get all owners."""
        raise NotImplementedError("unimplemented method")

    def validate(self, event: events.EventT):
        """Dispatch method for event validation methods."""
        handler: dict[type[events.Event],
            typing.Callable[[typing.Any], None]] = {
            events.ObjectCreateEvent: self._validate_object_create,
            events.ObjectUpdateEvent: self._validate_object_update,
            events.ObjectDeleteEvent: self._validate_object_delete,
            events.AnnotationCreateEvent: self._validate_annotation_create,
            events.AnnotationUpdateEvent: self._validate_annotation_update,
            events.AnnotationDeleteEvent: self._validate_annotation_delete,
            events.ReviewAcceptEvent: self._validate_review_accept,
            events.ReviewRejectEvent: self._validate_review_reject,
            events.OwnerAddEvent: self._validate_owner_add,
            events.OwnerRemoveEvent: self._validate_owner_remove,
        }

        if type(event) not in handler:
            raise NotImplementedError("unhandled event type in validate")

        handler[type(event)](event)

    @abc.abstractmethod
    def _validate_object_create(self, event: events.ObjectCreateEvent):
        """Validate :class:`gonk.core.events.ObjectCreateEvent`.

        Raises:
            ValidationError: Object UUID already exists.
            ValidationError: Schema name already in use.
            ValidationError: Duplicate hash detected.
            ValidationError: Object version not zero.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_object_update(self, event: events.ObjectUpdateEvent):
        """Validate :class:`gonk.core.events.ObjectUpdateEvent`.

        Raises:
            ValidationError: Object UUID does not exists.
            ValidationError: Duplicate hash detected.
            ValidationError: Schema name changed between versions.
            ValidationError: Incorrect object version.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_object_delete(self, event: events.ObjectDeleteEvent):
        """Validate :class:`gonk.core.events.ObjectDeleteEvent`.

        Raises:
            ValidationError: Object is a schema.
            ValidationError: Object identifier not found.
            ValidationError: Object is rejected.
            ValidationError: Object pending deletion.
            ValidationError: Object is deleted.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_annotation_create(self, event: events.AnnotationCreateEvent):
        """Validate :class:`gonk.core.events.AnnotationCreateEvent`.

        Raises:
            ValidationError: Annotation UUID already exists.
            ValidationError: Annotation version not zero.
            ValidationError: Object identifier not found.
            ValidationError: Annotating a rejected object.
            ValidationError: Annotating a deleted object.
            ValidationError: Annotating a schema.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_annotation_update(self, event: events.AnnotationUpdateEvent):
        """Validate :class:`gonk.core.events.AnnotationUpdateEvent`.

        Raises:
            ValidationError: Annotation UUID does not exists.
            ValidationError: Incorrect annotation version.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_annotation_delete(self, event: events.AnnotationDeleteEvent):
        """Validate :class:`gonk.core.events.AnnotationDeleteEvent`.

        Raises:
            ValidationError: Annotation identifier not found.
            ValidationError: Annotation is rejected.
            ValidationError: Annotation pending deletion.
            ValidationError: Annotation is deleted.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_review_accept(self, event: events.ReviewAcceptEvent):
        """Validate :class:`gonk.core.events.ReviewAcceptEvent`.

        Raises:
            ValidationError: Event not found.
            ValidationError: Event already reviewed.
            ValidationError: Review of non object or annotation event.
            ValidationError: Author is not owner.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_review_reject(self, event: events.ReviewRejectEvent):
        """Validate :class:`gonk.core.events.ReviewRejectEvent`.

        Raises:
            ValidationError: Event not found.
            ValidationError: Event already reviewed.
            ValidationError: Review of non object or annotation event.
            ValidationError: Author is not owner.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_owner_add(self, event: events.OwnerAddEvent):
        """Validate :class:`gonk.core.events.OwnerAddEvent`.

        Raises:
            ValidationError: Owner already exists.
            ValidationError: Author is not owner.
            ValidationError: First owner add event must be a self add.
        """
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _validate_owner_remove(self, event: events.OwnerRemoveEvent):
        """Validate :class:`gonk.core.events.OwnerRemoveEvent`.

        Raises:
            ValidationError: No owners to remove.
            ValidationError: Author is not owner.
            ValidationError: Target of remove event is not an owner.
            ValidationError: Cannot leave the dataset ownerless.
            ValidationError: Target is a higher rank than author.
        """
        raise NotImplementedError("unimplemented method")

    def consume(self, event: events.EventT):
        """Dispatch method for event consumption methods."""
        handler: dict[typing.Type[events.Event],
            typing.Callable[[typing.Any], None]] = {
            events.ObjectCreateEvent: self._consume_object_create,
            events.ObjectUpdateEvent: self._consume_object_update,
            events.ObjectDeleteEvent: self._consume_object_delete,
            events.AnnotationCreateEvent: self._consume_annotation_create,
            events.AnnotationUpdateEvent: self._consume_annotation_update,
            events.AnnotationDeleteEvent: self._consume_annotation_delete,
            events.ReviewAcceptEvent: self._consume_review_accept,
            events.ReviewRejectEvent: self._consume_review_reject,
            events.OwnerAddEvent: self._consume_owner_add,
            events.OwnerRemoveEvent: self._consume_owner_remove,
        }

        if type(event) not in handler:
            raise NotImplementedError("unhandled event type in consume")

        handler[type(event)](event)

    @abc.abstractmethod
    def _consume_object_create(self, event: events.ObjectCreateEvent):
        """Consume :class:`gonk.core.events.ObjectCreateEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_object_update(self, event: events.ObjectUpdateEvent):
        """Consume :class:`gonk.core.events.ObjectUpdateEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_object_delete(self, event: events.ObjectDeleteEvent):
        """Consume :class:`gonk.core.events.ObjectDeleteEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_annotation_create(self, event: events.AnnotationCreateEvent):
        """Consume :class:`gonk.core.events.AnnotationCreateEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_annotation_update(self, event: events.AnnotationUpdateEvent):
        """Consume :class:`gonk.core.events.AnnotationUpdateEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_annotation_delete(self, event: events.AnnotationDeleteEvent):
        """Consume :class:`gonk.core.events.AnnotationDeleteEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_review_accept(self, event: events.ReviewAcceptEvent):
        """Consume :class:`gonk.core.events.ReviewAcceptEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_review_reject(self, event: events.ReviewRejectEvent):
        """Consume :class:`gonk.core.events.ReviewRejectEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_owner_add(self, event: events.OwnerAddEvent):
        """Consume :class:`gonk.core.events.OwnerAddEvent`."""
        raise NotImplementedError("unimplemented method")

    @abc.abstractmethod
    def _consume_owner_remove(self, event: events.OwnerRemoveEvent):
        """Consume :class:`gonk.core.events.OwnerRemoveEvent`."""
        raise NotImplementedError("unimplemented method")




# Copyright 2023 - Compute Heavy Industries Incorporated
# This work is released, distributed, and licensed under AGPLv3.

import enum
import uuid
import json
import pathlib

from gonk.core import interfaces
from gonk.core import exceptions
from gonk.core import events

class RecordKeeper(interfaces.RecordKeeper):
    """File system backed RecordKeeper.

    Treats the event log as a linked list. Utilizes ``head`` and ``tail`` 
    files. Events are stored in depth=3 prefix-tree folder structure. For
    example, UUID ``96f76903-7b92-44d1-8e53-fc47a520684c`` would be stored
    in ``rk/events/9/6/f/``."""
    def __init__(self, parent_directory: pathlib.Path):
        super().__init__()
        if not parent_directory.exists():
            raise ValueError("parent directory does not exist")

        self.root_directory = parent_directory.joinpath("rk")
        self.event_directory = self.root_directory.joinpath("events")
        self.head_path = self.root_directory.joinpath("head")
        self.tail_path = self.root_directory.joinpath("tail")

        if not self.root_directory.exists():
            self.root_directory.mkdir()

        if not self.event_directory.exists():
            self.event_directory.mkdir()

    def _link_tail(self, key: str):
        if not self.tail_path.exists():
            return

        tail = self.tail_path.read_text()

        if len(tail) != 36:
            raise ValueError("invalid data for tail")

        record_path = self.event_directory.joinpath(
            f"{tail[0]}/{tail[1]}/{tail[2]}/{tail}")

        event_json = record_path.read_text()
        event_data = json.loads(event_json)

        if event_data["next"] is not None:
            raise ValueError("tail next is not none")

        event_data["next"] = key
        event_json = json.dumps(event_data)

        record_path.write_text(event_json)

    def add(self, event: events.EventT):
        event_data = event.serialize()
        event_data["type"] = event.__class__.__name__
        event_data["next"] = None

        event_json = json.dumps(event_data)

        key = event_data["uuid"]

        record_path = self.event_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}")
        record_path.mkdir(parents=True, exist_ok=True)
        record_path = record_path.joinpath(key)
        record_path.write_text(event_json)

        if not self.head_path.exists():
            self.head_path.write_text(key)

        self._link_tail(key)
        self.tail_path.write_text(key)

    def read(self, uuid_: uuid.UUID) -> events.Event:
        if not self.exists(uuid_):
            raise ValueError("event does not exist")

        key = str(uuid_)
        record_path = self.event_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}")
        event_json = record_path.read_text()
        event_data = json.loads(event_json)
        event = getattr(events, event_data["type"]).deserialize(event_data)

        return event

    def exists(self, uuid_: uuid.UUID) -> bool:
        key = str(uuid_)
        record_path = self.event_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}")
        return record_path.exists()

    def next(self, uuid_: uuid.UUID | None=None) -> uuid.UUID | None:
        if uuid_ is None:
            if not self.head_path.exists():
                return None

            head = self.head_path.read_text()

            if len(head) != 36:
                raise ValueError("invalid data for head")

            return uuid.UUID(head)

        if not self.exists(uuid_):
            raise ValueError("event does not exist")

        key = str(uuid_)
        record_path = self.event_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}")

        event_json = record_path.read_text()
        event_data = json.loads(event_json)
        next_ = event_data["next"]

        if next_ is None:
            return None

        return uuid.UUID(next_)

    def tail(self) -> uuid.UUID|None:
        if not self.tail_path.exists():
            return None

        tail = self.tail_path.read_text()

        return uuid.UUID(tail)

class ObjectStateT(enum.Enum):
    """Enum for Depot objects. """
    NONEXISTENT = 1<<0
    """Object does not exist."""
    READABLE = 1<<1
    """Object is readable."""
    WRITABLE = 1<<2
    """Object is writable."""

class Depot(interfaces.Depot):
    """File system backed Depot.

    Objects are stored in depth=3 prefix-tree folder structure. For
    example, UUID ``62af3937-4c5b-405b-a170-6ef12a5e246a`` would be stored
    in ``depot/6/2/a/``. When an object is still writable it will have the
    ``.wr`` extension. Upon finalization this extension is removed."""
    def __init__(self, parent_directory: pathlib.Path):
        super().__init__()
        if not parent_directory.exists():
            raise ValueError("parent directory does not exist")

        self.root_directory = parent_directory.joinpath("depot")

        if not self.root_directory.exists():
            self.root_directory.mkdir()

    def _state(self, identifier: events.Identifier) -> ObjectStateT:
        key = f"{identifier.uuid}.{identifier.version}"

        readable_path = self.root_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}")

        if readable_path.exists():
            return ObjectStateT.READABLE

        writable_path = self.root_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}.wr")

        if writable_path.exists():
            return ObjectStateT.WRITABLE

        return ObjectStateT.NONEXISTENT

    def exists(self, identifier: events.Identifier):
        return self._state(identifier) != ObjectStateT.NONEXISTENT

    def reserve(self, identifier: events.Identifier, size: int):
        if self._state(identifier) != ObjectStateT.NONEXISTENT:
            raise exceptions.StorageError(
                'identifier already exists in storage')

        key = f"{identifier.uuid}.{identifier.version}"

        object_path = self.root_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}")
        object_path.mkdir(parents=True, exist_ok=True)

        object_path = object_path.joinpath(f"{key}.wr")
        object_path.write_bytes(b'\x00'*size)

    def write(self, identifier: events.Identifier, offset: int, buf: bytes):
        state = self._state(identifier)
        if state == ObjectStateT.NONEXISTENT:
            raise exceptions.StorageError('identifier not found in storage')

        if state == ObjectStateT.READABLE:
            raise exceptions.StorageError('identifier already finalized')

        key = f"{identifier.uuid}.{identifier.version}"
        object_path = self.root_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}.wr")

        info = object_path.stat()

        if offset + len(buf) > info.st_size:
            raise exceptions.StorageError('write outside of reserved boundary')

        with object_path.open(mode="rb+") as f:
            f.seek(offset, 0)
            f.write(buf)

    def finalize(self, identifier: events.Identifier):
        state = self._state(identifier)
        if state == ObjectStateT.NONEXISTENT:
            raise exceptions.StorageError('identifier not found in storage')

        if state == ObjectStateT.READABLE:
            raise exceptions.StorageError('identifier already finalized')

        key = f"{identifier.uuid}.{identifier.version}"
        writable_path = self.root_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}.wr")

        readable_path = self.root_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}")

        writable_path.rename(readable_path)

    def read(self, identifier: events.Identifier, offset: int, size: int):
        state = self._state(identifier)
        if state == ObjectStateT.NONEXISTENT:
            raise exceptions.StorageError('identifier not found in storage')

        if state == ObjectStateT.WRITABLE:
            raise exceptions.StorageError('identifier not finalized')

        key = f"{identifier.uuid}.{identifier.version}"
        object_path = self.root_directory.joinpath(
            f"{key[0]}/{key[1]}/{key[2]}/{key}")

        with object_path.open(mode="rb") as f:
            f.seek(offset, 0)
            buf = f.read(size)

        return buf

    def purge(self, identifier: events.Identifier):
        state = self._state(identifier)
        if state == ObjectStateT.NONEXISTENT:
            raise exceptions.StorageError('identifier not found in storage')

        key = f"{identifier.uuid}.{identifier.version}"

        if state == ObjectStateT.READABLE:
            readable_path = self.root_directory.joinpath(
                f"{key[0]}/{key[1]}/{key[2]}/{key}")
            readable_path.unlink()

        if state == ObjectStateT.WRITABLE:
            writable_path = self.root_directory.joinpath(
                f"{key[0]}/{key[1]}/{key[2]}/{key}.wr")
            writable_path.unlink()



# Copyright 2023 - Compute Heavy Industries Incorporated
# This work is released, distributed, and licensed under AGPLv3.

import nacl
import typing
import hashlib

from nacl import signing

from gonk.core import interfaces
from gonk.core import exceptions
from gonk.core import events

class HashChainLinker:
    """Hash linker that links one event to the tail event in RecordKeeper."""
    def __init__(self, record_keeper: interfaces.RecordKeeper):
        self.record_keeper: interfaces.RecordKeeper = record_keeper
        """Dataset's RecordKeeper."""

    def link(self, event: events.EventT, author: str) -> events.EventT:
        """Calculate chained hash and populate 
            :class:`gonk.core.events.Event`'s 
            ``integrity`` and ``author`` fields."""
        tail = self.record_keeper.tail()
        prefix = b""
        if tail is not None:
            prev = self.record_keeper.read(tail)
            if prev.integrity is None:
                raise exceptions.ValidationError("tail event missing integrity")

            prefix = prev.integrity

        event.author = author
        event.integrity = hashlib.sha256(
            prefix + event.signature_bytes()).digest()
        return event

class HashChainValidator(interfaces.Validator):
    """Validator for hash linked events."""
    def __init__(self, record_keeper: interfaces.RecordKeeper):
        self.record_keeper = record_keeper
        """Dataset's RecordKeeper."""

    def validate(self, event: events.EventT):
        """Validate :class:`gonk.core.events.Event`'s 
            ``author`` and ``integrity`` fields."""
        if event.author is None:
            raise exceptions.ValidationError("event missing author")

        if event.integrity is None:
            raise exceptions.ValidationError("event missing integrity")
        
        tail = self.record_keeper.tail()
        prefix = b""
        if tail is not None:
            prev = self.record_keeper.read(tail)
            if prev.integrity is None:
                raise exceptions.ValidationError(
                    "previous event missing integrity")

            prefix = prev.integrity

        hash_ = hashlib.sha256(prefix + event.signature_bytes()).digest()
        if hash_ != event.integrity:
            raise exceptions.ValidationError(
                "event integrity failed to validate")

class Signer:
    """Event signer using nacl (ED25519)."""
    def __init__(self, signing_key: nacl.signing.SigningKey):
        self.signing_key: nacl.signing.SigningKey = signing_key
        """Key for signing."""
        self.verify_bytes: bytes = bytes(self.signing_key.verify_key)
        """Public key bytes for identity purposes."""

    def sign(self, event: events.EventT) -> events.EventT:
        """Sign event with ``signing_key``."""
        signed = self.signing_key.sign(event.signature_bytes())
        event.integrity = signed.signature
        event.author = self.verify_bytes.hex()
        return event

class SignatureValidator(interfaces.Validator):
    """Validator for signed events."""
    def __init__(self):
        super().__init__()

    def validate(self, event: events.EventT):
        """Validate that event is signed with the public key in ``author``."""
        if event.author is None:
            raise exceptions.ValidationError("event missing author")
        verify_key = nacl.signing.VerifyKey(bytes.fromhex(event.author))
        try:
            verify_key.verify(event.signature_bytes(), event.integrity)
        except nacl.exceptions.BadSignatureError as error:
            raise exceptions.ValidationError(
                "event integrity failed to validate") from error

class KeyPair:
    """Convenience class for serializing nacl key pairs."""
    def __init__(self, signing_key: bytes|None=None):
        if signing_key is not None:
            self.signing_key = nacl.signing.SigningKey(signing_key)
        else:
            self.signing_key = nacl.signing.SigningKey.generate()
        self.verify_key = self.signing_key.verify_key

    def serialize(self):
        return {
            "signing_key": bytes(self.signing_key).hex(),
            "verify_key": bytes(self.verify_key).hex(),
        }

    @classmethod
    def deserialize(cls, data: dict):
        return KeyPair(bytes.fromhex(data["signing_key"]))


# Copyright 2023 - Compute Heavy Industries Incorporated
# This work is released, distributed, and licensed under AGPLv3.

class ValidationError(Exception):
    """Event validation exception."""
    pass

class StorageError(Exception):
    """Exception for storage related issues."""
    pass


# Copyright 2023 - Compute Heavy Industries Incorporated
# This work is released, distributed, and licensed under AGPLv3.

import uuid
import enum
import typing
import struct
import datetime
import jsonschema

def tsnow() -> str:
    return f"{datetime.datetime.utcnow().isoformat('T')}Z"

### Enums ###
class ActionT(enum.Enum):
    """Enum for object and annotation event actions."""
    CREATE = 1<<0
    """Create event."""
    UPDATE = 1<<1
    """Update event."""
    DELETE = 1<<2
    """Delete event."""

class DecisionT(enum.Enum):
    """Enum for review event decisions."""
    ACCEPT = 1<<0
    """Event accepted."""
    REJECT = 1<<1
    """Event rejected."""

class OwnerActionT(enum.Enum):
    """Enum for owner event actions."""
    ADD = 1<<0
    """Owner added."""
    REMOVE = 1<<1
    """Owner removed."""

class HashTypeT(enum.Enum):
    """Enum for object and annotation hash types."""
    SHA256 = 1<<0
    """SHA256 hash is best hash."""

### Data Containers ###
class Identifier:
    """Identifies objects and annotations with a UUID and version number."""
    def __init__(self, uuid_: uuid.UUID, version: int):
        self.uuid: uuid.UUID = uuid_
        """Object or annotation's UUID."""
        self.version: int = version
        """Object or annotation's version."""

    def signature_bytes(self) -> bytes:
        """Return a byte-based representation for signing or hashing."""
        return b"".join([
            self.uuid.bytes,
            struct.pack("<Q", self.version),
        ])

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        return {
            "uuid": str(self.uuid),
            "version": self.version,
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        """Deserialize dictionary to instance."""
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(uuid.UUID(data["uuid"]), data["version"])

    @staticmethod
    def schema(relative="") -> dict:
        """Returns the JSON Schema for validating the serialized class.

        Args:
            relative: A base path for nested references.
        """
        return {
            "type": "object",
            "properties": {
                "uuid": {
                    "type": "string",
                    "format": "uuid",
                },
                "version": {
                    "type": "integer",
                    "minimum": 0,
                },
            },
            "required": [
                "uuid",
                "version",
            ],
        }

    def __hash__(self):
        return hash((self.uuid, self.version))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self.uuid == other.uuid and self.version == other.version

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return f"Identifier({self.uuid}, {self.version})"

class Object:
    """Metadata container for an object in the 
        :class:`gonk.core.interfaces.Depot`."""
    def __init__(self, name: str, format_: str, size: int, hash_type: HashTypeT,
        hash_: str, uuid_: typing.Optional[uuid.UUID] = None, version: int = 0):
        if uuid_ is None:
            uuid_ = uuid.uuid4()

        self.uuid: uuid.UUID = uuid_
        """Object's UUID."""
        self.version: int = version
        """Object's version."""
        self.name: str = name
        """Object's filename."""
        self.format: str = format_
        """Object's mimetype."""
        self.size: int = size
        """Object size in bytes."""
        self.hash_type: HashTypeT = hash_type
        """HashTypeT of the hash."""
        self.hash: str = hash_
        """Hex encoded hash of the object."""

    def identifier(self) -> Identifier:
        """Return the object identifier."""
        return Identifier(self.uuid, self.version)

    def signature_bytes(self) -> bytes:
        """Return a byte-based representation for signing or hashing."""
        return b"".join([
            self.uuid.bytes,
            struct.pack("<Q", self.version),
            self.name.encode(),
            self.format.encode(),
            struct.pack("<Q", self.size),
            struct.pack("<B", self.hash_type.value),
            bytes.fromhex(self.hash),
        ])

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        return {
            "uuid": str(self.uuid),
            "version": self.version,
            "name": self.name,
            "format": self.format,
            "size": self.size,
            "hash_type": self.hash_type.value,
            "hash": self.hash,
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        """Deserialize dictionary to instance."""
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            data["name"],
            data["format"],
            data["size"],
            HashTypeT(data["hash_type"]),
            data["hash"],
            uuid.UUID(data["uuid"]),
            data["version"])

    @staticmethod
    def schema(relative="") -> dict:
        """Returns the JSON Schema for validating the serialized class.

        Args:
            relative: A base path for nested references.
        """
        return {
            "type": "object",
            "properties": {
                "uuid": {
                    "type": "string",
                    "format": "uuid",
                },
                "version": {
                    "type": "integer",
                    "minimum": 0,
                },
                "name": {
                    "type": "string"
                },
                "format": {
                    "type": "string"
                },
                "size": {
                    "type": "integer",
                    "minimum": 0,
                },
                "hash_type": {
                    "type": "integer"
                },
                "hash": {
                    "type": "string",
                    "minLength": 64,
                    "maxLength": 64,
                    "pattern": "^[0-9a-fA-F]{64}$"
                }
            },
            "required": [
                "uuid",
                "version",
                "name",
                "format",
                "size",
                "hash",
                "hash_type",
            ],
        }

    def __copy__(self):
        return Object(self.name, self.format, self.size, self.hash_type,
          self.hash, self.uuid, self.version)

    def __eq__(self, other):
        return self.uuid == other.uuid and \
            self.version == other.version and \
            self.name == other.name and \
            self.format == other.format and \
            self.size == other.size and \
            self.hash_type == other.hash_type and \
            self.hash == other.hash

    def __ne__(self, other):
        return not self.__eq__(other)

class Annotation:
    """Metadata container for an annotation in the 
        :class:`gonk.core.interfaces.Depot`."""
    def __init__(self, schema_: Identifier, size: int, hash_type: HashTypeT,
        hash_: str, uuid_: typing.Optional[uuid.UUID] = None, version: int = 0):
        if uuid_ is None:
            uuid_ = uuid.uuid4()

        self.uuid: uuid.UUID = uuid_ 
        """Annotation's UUID."""
        self.version: int = version
        """Annotation's version."""
        self.schema_: Identifier = schema_
        """Schema identifier."""
        self.size: int = size
        """Annotation size in bytes."""
        self.hash_type: HashTypeT = hash_type
        """Annotation's hash type."""
        self.hash: str = hash_
        """Hex encoded hash."""

    def identifier(self) -> Identifier:
        """Return the identifier for this annotation."""
        return Identifier(self.uuid, self.version)

    def signature_bytes(self) -> bytes:
        """Return a byte-based representation for signing or hashing."""
        return b"".join([
            self.uuid.bytes,
            struct.pack("<Q", self.version),
            self.schema_.signature_bytes(),
            struct.pack("<Q", self.size),
            struct.pack("<B", self.hash_type.value),
            bytes.fromhex(self.hash),
        ])

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        return {
            "uuid": str(self.uuid),
            "version": self.version,
            "schema": self.schema_.serialize(),
            "size": self.size,
            "hash_type": self.hash_type.value,
            "hash": self.hash,
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        """Deserialize dictionary to instance."""
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            Identifier.deserialize(data["schema"]),
            data["size"],
            HashTypeT(data["hash_type"]),
            data["hash"],
            uuid.UUID(data["uuid"]),
            data["version"])

    @staticmethod
    def schema(relative="") -> dict:
        """Returns the JSON Schema for validating the serialized class.

        Args:
            relative: A base path for nested references.
        """
        return {
            "type": "object",
            "definitions": {
                "identifier": Identifier.schema("/definitions/identifier"),
            },
            "properties": {
                "uuid": {
                    "type": "string",
                    "format": "uuid",
                },
                "version": {
                    "type": "integer",
                    "minimum": 0,
                },
                "schema": {
                    "$ref": f"#{relative}/definitions/identifier",
                },
                "size": {
                    "type": "integer",
                    "minimum": 0,
                },
                "hash_type": {
                    "type": "integer"
                },
                "hash": {
                    "type": "string",
                    "minLength": 64,
                    "maxLength": 64,
                    "pattern": "^[0-9a-fA-F]{64}$"
                }
            },
            "required": [
                "uuid",
                "version",
                "schema",
                "size",
                "hash",
                "hash_type",
            ],
        }

    def __eq__(self, other):
        return self.uuid == other.uuid and \
            self.version == other.version and \
            self.schema_ == other.schema_ and \
            self.size == other.size and \
            self.hash_type == other.hash_type and \
            self.hash == other.hash

    def __ne__(self, other):
        return not self.__eq__(other)

    def __copy__(self):
        return Annotation(self.schema_, self.size, self.hash_type, self.hash,
            self.uuid, self.version)

### Events ###
class Event:
    """Parent class for all event types."""
    def __init__(self,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        if uuid_ is None:
            uuid_ = uuid.uuid4()

        if timestamp is None:
            timestamp = tsnow()

        self.uuid: uuid.UUID = uuid_
        """Event's UUID."""
        self.timestamp: str = timestamp
        """Event's timestamp."""
        self.integrity: typing.Optional[bytes] = integrity
        """Byte field for hashes or signatures."""
        self.author: typing.Optional[str] = author
        """String field for the event's author."""
    def signature_bytes(self) -> bytes:
        """Return a byte-based representation for signing or hashing."""
        return b"".join([
            self.uuid.bytes,
            self.timestamp.encode(),
        ])

    def serialize(self) -> dict:
        """Serialize instance to dictionary."""
        integrity = None
        if self.integrity is not None:
            integrity = self.integrity.hex()

        return {
            "uuid": str(self.uuid),
            "timestamp": self.timestamp,
            "integrity": integrity,
            "author": self.author,
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        """Deserialize dictionary to instance."""
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        """Returns the JSON Schema for validating the serialized class.

        Args:
            relative: A base path for nested references.
        """
        return {
            "type": "object",
            "properties": {
                "uuid": {
                    "type": "string",
                    "format": "uuid",
                },
                "timestamp": {
                    "type": "string"
                },
                "integrity": {
                    "type": "string",
                    "minLength": 32,
                    "maxLength": 256,
                    "pattern": "^[0-9a-fA-F]{32,256}$"
                },
                "author": {
                    "type": "string",
                    "minLength": 1,
                    "maxLength": 512,
                },
            },
            "required": [
                "uuid",
                "timestamp",
                "integrity",
                "author",
            ],
        }

    def __eq__(self, other):
        return self.uuid == other.uuid and \
            self.timestamp == other.timestamp and \
            self.integrity == other.integrity and \
            self.author == other.author

    def __ne__(self, other):
        return not self.__eq__(other)

EventT = typing.TypeVar("EventT", bound=Event)

### Object Events ###
class ObjectEvent(Event):
    """Parent class for object-specific events (create, delete, update)."""
    def __init__(self,
        action: ActionT,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(uuid_, timestamp, integrity, author)
        self.action: ActionT = action
        """The event action."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            struct.pack("<B", self.action.value),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "action": self.action.value,
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(ActionT(data["action"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "event": Event.schema("/definitions/event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/event" }
            ],
            "properties": {
                "action": {
                    "type": "integer"
                },
            },
            "required": [
                "action",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.action == other.action

    def __ne__(self, other):
        return not self.__eq__(other)

class ObjectCreateEvent(ObjectEvent):
    """Event used for object creation."""
    def __init__(self,
        object_: Object,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):
        super().__init__(ActionT.CREATE, uuid_, timestamp, integrity, author)

        self.object: Object = object_
        """The :class:`Object` being created."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            self.object.signature_bytes(),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "object": self.object.serialize(),
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(Object.deserialize(data["object"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "object": Object.schema("/definitions/object"),
                "object_event": ObjectEvent.schema("/definitions/object_event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/object_event" }
            ],
            "properties": {
                "object": {
                    "$ref": f"#{relative}/definitions/object",
                }
            },
            "required": [
                "object",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.object == other.object

    def __ne__(self, other):
        return not self.__eq__(other)

class ObjectUpdateEvent(ObjectEvent):
    """Event used for object updates."""
    def __init__(self,
        object_: Object,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):
        
        super().__init__(ActionT.UPDATE, uuid_, timestamp, integrity, author)
        self.object: Object = object_
        """The updated version of the :class:`Object`."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            self.object.signature_bytes(),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "object": self.object.serialize(),
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(Object.deserialize(data["object"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "object": Object.schema("/definitions/object"),
                "object_event": ObjectEvent.schema("/definitions/object_event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/object_event" }
            ],
            "properties": {
                "object": {
                    "$ref": f"#{relative}/definitions/object",
                }
            },
            "required": [
                "object",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.object == other.object

    def __ne__(self, other):
        return not self.__eq__(other)

class ObjectDeleteEvent(ObjectEvent):
    """Event used for object deletion."""
    def __init__(self,
        object_identifier: Identifier,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(ActionT.DELETE, uuid_, timestamp, integrity, author)
        self.object_identifier: Identifier = object_identifier
        """The deleted :class:`Object`'s :class:`Identifier`."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            self.object_identifier.signature_bytes(),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "object_identifier": self.object_identifier.serialize(),
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            Identifier.deserialize(data["object_identifier"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "identifier": Identifier.schema("/definitions/identifier"),
                "object_event": ObjectEvent.schema("/definitions/object_event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/object_event" }
            ],
            "properties": {
                "object_identifier": {
                    "$ref": f"#{relative}/definitions/identifier",
                }
            },
            "required": [
                "object_identifier",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.object_identifier == other.object_identifier

    def __ne__(self, other):
        return not self.__eq__(other)

### Annotation Events ###
class AnnotationEvent(Event):
    """Parent class for annotation-specific events (create, update, delete)."""
    def __init__(self,
        action: ActionT,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(uuid_, timestamp, integrity, author)
        self.action: ActionT = action
        """The event action."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            struct.pack("<B", self.action.value),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "action": self.action.value,
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(ActionT(data["action"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "event": Event.schema("/definitions/event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/event" }
            ],
            "properties": {
                "action": {
                    "type": "integer"
                },
            },
            "required": [
                "action",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.action == other.action

    def __ne__(self, other):
        return not self.__eq__(other)

class AnnotationCreateEvent(AnnotationEvent):
    """Event used for creating an annotation."""
    def __init__(self,
        object_identifiers: list[Identifier],
        annotation: Annotation,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(ActionT.CREATE, uuid_, timestamp, integrity, author)
        self.object_identifiers: list[Identifier] = object_identifiers
        """Object :class:`Identifier`\s that this annotation references."""
        self.annotation: Annotation = annotation
        """The :class:`Annotation` being created."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            *[ea.signature_bytes() for ea in self.object_identifiers],
            self.annotation.signature_bytes(),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "annotation": self.annotation.serialize(),
            "object_identifiers": [ea.serialize() 
                for ea in self.object_identifiers],
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            [Identifier.deserialize(ea) for ea in data["object_identifiers"]],
            Annotation.deserialize(data["annotation"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "annotation": Annotation.schema("/definitions/annotation"),
                "annotation_event":
                    ObjectEvent.schema("/definitions/annotation_event"),
                "identifier": Identifier.schema("/definitions/identifier"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/annotation_event" }
            ],
            "properties": {
                "annotation": {
                    "$ref": f"#{relative}/definitions/annotation",
                },
                "object_identifiers": {
                    "type": "array",
                    "items": {
                        "$ref": f"#{relative}/definitions/identifier",
                    },
                },
            },
            "required": [
                "annotation",
                "object_identifiers",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.object_identifiers == other.object_identifiers and \
            self.annotation == other.annotation

    def __ne__(self, other):
        return not self.__eq__(other)

class AnnotationUpdateEvent(AnnotationEvent):
    """Event used for updating an annotation."""
    def __init__(self,
        annotation: Annotation,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(ActionT.UPDATE, uuid_, timestamp, integrity, author)
        self.annotation = annotation
        """The :class:`Annotation` being updated."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            self.annotation.signature_bytes(),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "annotation": self.annotation.serialize(),
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            Annotation.deserialize(data["annotation"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "annotation": Annotation.schema("/definitions/annotation"),
                "annotation_event":
                    ObjectEvent.schema("/definitions/annotation_event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/annotation_event" }
            ],
            "properties": {
                "annotation": {
                    "$ref": f"#{relative}/definitions/annotation",
                },
            },
            "required": [
                "annotation",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.annotation == other.annotation

    def __ne__(self, other):
        return not self.__eq__(other)

class AnnotationDeleteEvent(AnnotationEvent):
    """Event used for deleting an annotation."""
    def __init__(self,
        annotation_identifier: Identifier,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(ActionT.DELETE, uuid_, timestamp, integrity, author)
        self.annotation_identifier: Identifier = annotation_identifier
        """The :class:`Identifier` of the annotation being deleted."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            self.annotation_identifier.signature_bytes(),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "annotation_identifier": self.annotation_identifier.serialize(),
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            Identifier.deserialize(data["annotation_identifier"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "identifier": Identifier.schema("/definitions/identifier"),
                "annotation_event":
                    AnnotationEvent.schema("/definitions/annotation_event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/annotation_event" }
            ],
            "properties": {
                "annotation_identifier": {
                    "$ref": f"#{relative}/definitions/identifier",
                }
            },
            "required": [
                "annotation_identifier",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.annotation_identifier == other.annotation_identifier

    def __ne__(self, other):
        return not self.__eq__(other)

### Review Events ###
class ReviewEvent(Event):
    """Parent class for review events."""
    def __init__(self,
        decision: DecisionT,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(uuid_, timestamp, integrity, author)
        self.decision: DecisionT = decision
        """The decision of the review (accept, reject)."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            struct.pack("<B", self.decision.value),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "decision": self.decision.value,
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            DecisionT(data["decision"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "event": Event.schema("/definitions/event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/event" }
            ],
            "properties": {
                "decision": {
                    "type": "integer"
                },
            },
            "required": [
                "decision",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.decision == other.decision

    def __ne__(self, other):
        return not self.__eq__(other)

class ReviewAcceptEvent(ReviewEvent):
    """Event used for accepting pending object and annotation events."""
    def __init__(self,
        event_uuid: uuid.UUID,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(DecisionT.ACCEPT, uuid_, timestamp, integrity, author)
        self.event_uuid: uuid.UUID = event_uuid
        """The change being accepted."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            self.event_uuid.bytes,
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "event_uuid": str(self.event_uuid),
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            uuid.UUID(data["event_uuid"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "review_event": ReviewEvent.schema("/definitions/review_event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/review_event" }
            ],
            "properties": {
                "event_uuid": {
                    "type": "string",
                    "format": "uuid",
                },
            },
            "required": [
                "event_uuid",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.event_uuid == other.event_uuid

    def __ne__(self, other):
        return not self.__eq__(other)

class ReviewRejectEvent(ReviewEvent):
    """Event used for rejecting pending object and annotation events."""
    def __init__(self,
        event_uuid: uuid.UUID,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):
    
        super().__init__(DecisionT.REJECT, uuid_, timestamp, integrity, author)
        self.event_uuid: uuid.UUID = event_uuid
        """The change being rejected."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            self.event_uuid.bytes,
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "event_uuid": str(self.event_uuid),
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            uuid.UUID(data["event_uuid"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "review_event": ReviewEvent.schema("/definitions/review_event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/review_event" }
            ],
            "properties": {
                "event_uuid": {
                    "type": "string",
                    "format": "uuid",
                },
            },
            "required": [
                "event_uuid",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.event_uuid == other.event_uuid

    def __ne__(self, other):
        return not self.__eq__(other)

### Ownership Events ###
class OwnerEvent(Event):
    """Parent class for owner events."""
    def __init__(self,
        owner: str,
        owner_action: OwnerActionT,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):

        super().__init__(uuid_, timestamp, integrity, author)
        self.owner: str = owner
        """The owner affected."""
        self.owner_action: OwnerActionT = owner_action
        """Whether the owner is being added or removed."""

    def signature_bytes(self) -> bytes:
        return b"".join([
            super().signature_bytes(),
            self.owner.encode(),
            struct.pack("<B", self.owner_action.value),
        ])

    def serialize(self) -> dict:
        return super().serialize() | {
            "owner": self.owner,
            "owner_action": self.owner_action.value,
        }

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            data["owner"],
            OwnerActionT(data["owner_action"]),
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

    @staticmethod
    def schema(relative="") -> dict:
        return {
            "type": "object",
            "definitions": {
                "event": Event.schema("/definitions/event"),
            },
            "allOf": [
                { "$ref": f"#{relative}/definitions/event" }
            ],
            "properties": {
                "owner_action": {
                    "type": "integer"
                },
                "owner": {
                    "type": "string",
                    "minLength": 1,
                    "maxLength": 512,
                },
            },
            "required": [
                "owner",
                "owner_action",
            ],
        }

    def __eq__(self, other):
        return super().__eq__(other) and \
            self.owner == other.owner and \
            self.owner_action == other.owner_action

    def __ne__(self, other):
        return not self.__eq__(other)

class OwnerAddEvent(OwnerEvent):
    """Event used for adding an owner."""
    def __init__(self,
        owner: str,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):
        super().__init__(
            owner,
            OwnerActionT.ADD,
            uuid_,
            timestamp,
            integrity,
            author)

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            data["owner"],
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])

class OwnerRemoveEvent(OwnerEvent):
    """Event used for removing an owner."""
    def __init__(self,
        owner: str,
        uuid_: typing.Optional[uuid.UUID] = None,
        timestamp: typing.Optional[str] = None,
        integrity: typing.Optional[bytes] = None,
        author: typing.Optional[str] = None):
        super().__init__(
            owner,
            OwnerActionT.REMOVE,
            uuid_,
            timestamp,
            integrity,
            author)

    @classmethod
    def deserialize(cls, data: dict) -> typing.Self:
        jsonschema.validate(instance=data, schema=cls.schema())
        return cls(
            data["owner"],
            uuid.UUID(data["uuid"]),
            data["timestamp"],
            bytes.fromhex(data["integrity"]),
            data["author"])


# Copyright 2023 - Compute Heavy Industries Incorporated
# This work is released, distributed, and licensed under AGPLv3.

import json
import uuid
import typing
import jsonschema

from gonk.core import interfaces
from gonk.core import exceptions
from gonk.core import events

def is_schema(name: str) -> bool:
    return name.startswith("schema-")

class FieldValidator(interfaces.Validator):
    """Validator for the fields of :class:`gonk.core.events.Object` and 
        :class:`gonk.core.events.Annotation`."""
    def validate(self, event: events.EventT):
        handler: dict[typing.Type[events.Event],
            typing.Callable[[typing.Any], None]] = {
            events.ObjectCreateEvent: self._validate_object,
            events.ObjectUpdateEvent: self._validate_object,
            events.AnnotationCreateEvent: self._validate_annotation,
            events.AnnotationUpdateEvent: self._validate_annotation,
        }

        if type(event) not in handler:
            return

        handler[type(event)](event)

    def _validate_object(self,
        event: events.ObjectCreateEvent|events.ObjectUpdateEvent):

        object_ = event.object
        if not isinstance(object_.uuid, uuid.UUID):
            raise exceptions.ValidationError("UUID is not of type uuid.UUID")

        if object_.version < 0:
            raise exceptions.ValidationError(
                "version must be a non-negative integer")

        if len(object_.name) == 0:
            raise exceptions.ValidationError("object name cannot be empty")

        if len(object_.format) == 0:
            raise exceptions.ValidationError("object format cannot be empty")

        if object_.size < 0:
            raise exceptions.ValidationError(
                "size must be a non-negative integer")

        if object_.hash_type != events.HashTypeT.SHA256:
            raise exceptions.ValidationError("hash type must be SHA256")

        if len(object_.hash) != 64:
            raise exceptions.ValidationError(
                "hash should be a hex encoded SHA256")

    def _validate_annotation(self,
        event: events.AnnotationCreateEvent|events.AnnotationUpdateEvent):
        annotation = event.annotation
        if not isinstance(annotation.uuid, uuid.UUID):
            raise exceptions.ValidationError("UUID is not of type uuid.UUID")

        if annotation.version < 0:
            raise exceptions.ValidationError(
                "version must be a positive integer")

        if annotation.size < 0:
            raise exceptions.ValidationError(
                "size must be a non-negative integer")

        if not isinstance(annotation.schema_, events.Identifier):
            raise exceptions.ValidationError("schema must be an identifier")

        if annotation.hash_type != events.HashTypeT.SHA256:
            raise exceptions.ValidationError("hash type must be SHA256")

        if len(annotation.hash) != 64:
            raise exceptions.ValidationError(
                "hash should be a hex encoded SHA256")

class SchemaValidator(interfaces.Validator, interfaces.Consumer):
    """JSON Schema validator to ensure schemas are valid and 
    that annotations conform to the schema they reference. Schema objects 
    must be mimetype ``application/schema+json``."""
    def __init__(self, depot: interfaces.Depot):
        super().__init__()
        self.depot: interfaces.Depot = depot
        self.schemas: set[events.Identifier] = set()

    def validate(self, event: events.EventT):
        handler: dict[typing.Type[events.Event],
            typing.Callable[[typing.Any], None]] = {
            events.ObjectCreateEvent: self._validate_object_create,
            events.ObjectUpdateEvent: self._validate_object_update,
            events.AnnotationCreateEvent: self._validate_annotation_create,
            events.AnnotationUpdateEvent: self._validate_annotation_update,
        }

        if type(event) not in handler:
            return

        handler[type(event)](event)

    def _validate_object(self, object_):
        if not is_schema(object_.name):
            return

        if object_.format != "application/schema+json":
            return

        bs = self.depot.read(object_.identifier(), 0, object_.size)
        schema = json.loads(bs.decode())
        try:
            jsonschema.protocols.Validator.check_schema(schema)
        except jsonschema.exceptions.SchemaError as error:
            raise exceptions.ValidationError("invalid JSON schema") from error

    def _validate_annotation(self, annotation):
        if annotation.schema_ not in self.schemas:
            return

        schema_bs = bytes()
        off = 0
        chunk = 1024*5
        while True:
            buf = self.depot.read(annotation.schema_, off, chunk)
            schema_bs += buf
            off += len(buf)
            if len(buf) < chunk:
                break

        schema = json.loads(schema_bs.decode())

        annotation_bs = self.depot.read(
            annotation.identifier(), 0, annotation.size)
        instance = json.loads(annotation_bs.decode())

        try:
            jsonschema.validate(instance, schema)
        except jsonschema.exceptions.ValidationError as error:
            raise exceptions.ValidationError(
                "annotation does not match schema") from error

    def _validate_object_create(self, event: events.ObjectCreateEvent):
        self._validate_object(event.object)

    def _validate_object_update(self, event: events.ObjectCreateEvent):
        self._validate_object(event.object)

    def _validate_annotation_create(
        self, event: events.AnnotationCreateEvent):
        self._validate_annotation(event.annotation)

    def _validate_annotation_update(
        self, event: events.AnnotationUpdateEvent):
        self._validate_annotation(event.annotation)

    def consume(self, event: events.EventT):
        handler: dict[typing.Type[events.Event],
            typing.Callable[[typing.Any], None]] = {
            events.ObjectCreateEvent: self._consume_object_create,
            events.ObjectUpdateEvent: self._consume_object_update,
        }

        if type(event) not in handler:
            return

        handler[type(event)](event)

    def _consume_object(self, object_: events.Object):
        if not is_schema(object_.name):
            return

        if object_.format != "application/schema+json":
            return

        self.schemas.add(object_.identifier())

    def _consume_object_create(self, event: events.ObjectCreateEvent):
        self._consume_object(event.object)

    def _consume_object_update(self, event: events.ObjectUpdateEvent):
        self._consume_object(event.object)


# Copyright 2023 - Compute Heavy Industries Incorporated
# This work is released, distributed, and licensed under AGPLv3.

import uuid
import json
import enum
import typing
import pathlib
import sqlite3

from gonk.core import interfaces
from gonk.core import exceptions
from gonk.core import validators
from gonk.core import events

class RecordKeeper(interfaces.RecordKeeper):
    """SQLite backed RecordKeeper.

    Events are stored in a table, JSON serialized. ID primary key and
    event UUID are indexed."""
    def __init__(self, parent_directory: pathlib.Path):
        super().__init__()
        if not parent_directory.exists():
            raise ValueError("parent directory does not exist")

        self.database_path = parent_directory.joinpath("rk.db")

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.executescript("""
            CREATE TABLE IF NOT EXISTS events (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                uuid TEXT NOT NULL,
                event JSON NOT NULL
            );

            CREATE INDEX idx_events_uuid ON events(uuid);
        """)
        con.commit()
        con.close()

    def add(self, event: events.EventT):
        event_data = event.serialize()
        event_data["type"] = event.__class__.__name__

        event_json = json.dumps(event_data)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(
            "INSERT INTO events (uuid, event) VALUES (?, ?)",
            (event_data["uuid"], event_json))
        con.commit()
        con.close()

    def read(self, uuid_: uuid.UUID) -> events.Event:
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("SELECT event FROM events WHERE uuid = ?", (str(uuid_),))

        res = cur.fetchone()
        con.close()
        if res is None:
            raise ValueError("event does not exist")

        event_json, = res
        event_data = json.loads(event_json)
        event = getattr(events, event_data["type"]).deserialize(event_data)

        return event

    def exists(self, uuid_: uuid.UUID) -> bool:
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("SELECT id FROM events WHERE uuid = ?", (str(uuid_),))

        res = cur.fetchone()
        con.close()
        if res is None:
            return False

        return True

    def next(self, uuid_: uuid.UUID | None=None) -> uuid.UUID | None:
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        if uuid_ is None:
            cur.execute("SELECT uuid FROM events ORDER BY id LIMIT 1")
            res = cur.fetchone()

            if res is None:
                con.close()
                return None

            next_, = res
            con.close()
            return uuid.UUID(next_)

        cur.execute("SELECT id FROM events WHERE uuid = ?", (str(uuid_),))
        res = cur.fetchone()

        if res is None:
            con.close()
            return None

        id_, = res
        cur.execute("""SELECT uuid
                FROM events
                WHERE id > ?
                ORDER BY id
                LIMIT 1""",
            (id_,))
        res = cur.fetchone()

        if res is None:
            con.close()
            return None

        next_, = res
        con.close()
        return uuid.UUID(next_)

    def tail(self) -> uuid.UUID | None:
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("SELECT uuid FROM events ORDER BY id DESC LIMIT 1")
        res = cur.fetchone()

        if res is None:
            con.close()
            return None

        tail, = res
        con.close()
        return uuid.UUID(tail)

class StatusT(enum.Enum):
    """Enum for tracking object and annotation status in state.

    No status implies create accapted."""
    CREATE_PENDING = 1<<0
    """Create event pending."""
    CREATE_REJECTED = 1<<1
    """Create event rejected."""
    DELETE_PENDING = 1<<2
    """Delete event pending."""
    DELETE_ACCEPTED = 1<<3
    """Delete event accepted."""

class State(interfaces.State):
    """SQLite backed State.

    See parent :class:`gonk.core.interfaces.State` for more information."""
    def __init__(self,
        parent_directory: pathlib.Path,
        record_keeper: interfaces.RecordKeeper):

        super().__init__()
        if not parent_directory.exists():
            raise ValueError("parent directory does not exist")

        self.record_keeper = record_keeper

        self.database_path = parent_directory.joinpath("state.db")

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.executescript("""
            CREATE TABLE IF NOT EXISTS events (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                uuid TEXT NOT NULL,
                type TEXT NOT NULL
            );

            CREATE TABLE IF NOT EXISTS event_review_link (
                event_uuid TEXT NOT NULL,
                review_uuid TEXT NOT NULL,
                accepted BOOL NOT NULL
            );

            CREATE TABLE IF NOT EXISTS objects (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                uuid TEXT NOT NULL,
                version INTEGER NOT NULL,
                object JSON NOT NULL
            );

            CREATE TABLE IF NOT EXISTS schemas (
                name TEXT NOT NULL,
                uuid TEXT NOT NULL,
                version INTEGER NOT NULL
            );

            CREATE TABLE IF NOT EXISTS object_event_link (
                object_uuid TEXT NOT NULL,
                object_version INTEGER NOT NULL,
                event_uuid TEXT NOT NULL
            );

            CREATE TABLE IF NOT EXISTS object_status (
                uuid TEXT NOT NULL,
                version INTEGER NOT NULL,
                status TEXT NOT NULL
            );

            CREATE TABLE IF NOT EXISTS annotations (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                uuid TEXT NOT NULL,
                version INTEGER NOT NULL,
                annotation JSON NOT NULL
            );

            CREATE TABLE IF NOT EXISTS object_annotation_link (
                object_uuid TEXT NOT NULL,
                object_version INTEGER NOT NULL,
                annotation_uuid TEXT NOT NULL
            );

            CREATE TABLE IF NOT EXISTS annotation_event_link (
                annotation_uuid TEXT NOT NULL,
                annotation_version INTEGER NOT NULL,
                event_uuid TEXT NOT NULL
            );

            CREATE TABLE IF NOT EXISTS annotation_status (
                uuid TEXT NOT NULL,
                version INTEGER NOT NULL,
                status TEXT NOT NULL
            );

            CREATE TABLE IF NOT EXISTS owners (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                owner TEXT NOT NULL
            );
        """)

        cur.execute("""SELECT name 
            FROM sqlite_master 
            WHERE type='index'""")

        res = cur.fetchall()

        indices = {idx for idx, in res}

        statements = {
            "idx_events_uuid": 
                """CREATE INDEX idx_events_uuid 
                    ON events(uuid)""",
            "idx_objects_uuid": 
                """CREATE INDEX idx_objects_uuid 
                    ON objects(uuid)""",
            "idx_objects_object_hash": 
                """CREATE INDEX idx_objects_object_hash 
                    ON objects(object->>'$.hash')""",
            "idx_object_status_uuid": 
                """CREATE INDEX idx_object_status_uuid 
                    ON object_status(uuid)""",
            "idx_object_event_link_object_uuid": 
                """CREATE INDEX idx_object_event_link_object_uuid 
                    ON object_event_link(object_uuid)""",
            "idx_object_event_link_event_uuid": 
                """CREATE INDEX idx_object_event_link_event_uuid 
                    ON object_event_link(event_uuid)""",
            "idx_annotations_uuid": 
                """CREATE INDEX idx_annotations_uuid 
                    ON annotations(uuid)""",
            "idx_annotation_status_uuid": 
                """CREATE INDEX idx_annotation_status_uuid 
                    ON annotation_status(uuid)""",
        }

        for name, statement in statements.items():
            if name not in indices:
                cur.execute(statement)

        con.commit()
        con.close()

    def _accepted_to_review(self, type_, accepted):
        reviewables = [
            "ObjectCreateEvent", 
            "ObjectUpdateEvent", 
            "ObjectDeleteEvent",
            "AnnotationCreateEvent",
            "AnnotationUpdateEvent",
            "AnnotationDeleteEvent"]

        review = None
        if accepted is None:
            if type_ in reviewables:
                review = "PENDING"
        else:
            if bool(accepted):
                review = "ACCEPTED"
            else:
                review = "REJECTED"

        return review

    def events_by_object(self, identifier: events.Identifier):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT E.uuid, E.type, ERL.accepted 
                FROM events E
                INNER JOIN object_event_link OEL
                    ON E.uuid = OEL.event_uuid
                LEFT JOIN event_review_link ERL
                    ON E.uuid = ERL.event_uuid
                WHERE OEL.object_uuid = ?
                    AND OEL.object_version = ?
                ORDER BY E.id""",
            (str(identifier.uuid), identifier.version))

        res = cur.fetchall()
        con.close()

        return [interfaces.EventInfo(uuid.UUID(uu), type_, 
                self._accepted_to_review(type_, accepted)) 
            for uu, type_, accepted in res]

    def events_by_annotation(self, identifier: events.Identifier):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT E.uuid, E.type, ERL.accepted
                FROM events E
                INNER JOIN annotation_event_link AEL
                    ON E.uuid = AEL.event_uuid
                LEFT JOIN event_review_link ERL
                    ON E.uuid = ERL.event_uuid
                WHERE AEL.annotation_uuid = ?
                    AND AEL.annotation_version = ?
                ORDER BY E.id""",
            (str(identifier.uuid), identifier.version))

        res = cur.fetchall()
        con.close()

        return [interfaces.EventInfo(uuid.UUID(uu), type_, 
                self._accepted_to_review(type_, accepted)) 
            for uu, type_, accepted in res]

    def events_all(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where += """ WHERE E.id > (
                SELECT id 
                FROM events 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT E.uuid, E.type, ERL.accepted
            FROM events E
            LEFT JOIN event_review_link ERL
                ON E.uuid = ERL.event_uuid
            {where} 
            ORDER BY E.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [interfaces.EventInfo(uuid.UUID(uu), type_, 
                self._accepted_to_review(type_, accepted))
            for uu, type_, accepted in res]

    def annotations_all(self, 
        uuid_: None|uuid.UUID = None, after: None|uuid.UUID = None):

        if after is not None and uuid_ is not None:
            raise ValueError("only provide one of uuid_, after")

        params: tuple = tuple()
        where = ""
        if uuid_ is not None:
            where += "WHERE A.uuid = ?"
            params = (str(uuid_),)

        if after is not None:
            where += """WHERE A.id > (
                SELECT id 
                FROM annotations 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT DISTINCT A.uuid, 
                COUNT(A.version) OVER (PARTITION BY A.uuid),
                FIRST_VALUE(A.id) OVER (PARTITION BY A.uuid)
            FROM annotations A
            {where} 
            ORDER BY A.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [interfaces.AnnotationInfo(
            uuid.UUID(uu), vers) for uu, vers, _ in res]

    def annotations_by_object(self, object_identifier: events.Identifier):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT DISTINCT OAL.annotation_uuid,
                    COUNT(A.version) OVER (PARTITION BY A.uuid)
                FROM object_annotation_link OAL
                INNER JOIN annotations A
                    ON A.uuid = OAL.annotation_uuid
                WHERE OAL.object_uuid = ?
                    AND OAL.object_version = ?""", 
            (str(object_identifier.uuid), object_identifier.version))

        res = cur.fetchall()
        con.close()

        return [interfaces.AnnotationInfo(
            uuid.UUID(uu), vers) for uu, vers in res]

    def annotations_by_status(self, status: str, after: None|uuid.UUID = None):
        if status == "pending":
            return self._annotations_pending(after)
        elif status == "accepted":
            return self._annotations_accepted(after)
        elif status == "deleted":
            return self._annotations_deleted(after)
        elif status == "rejected":
            return self._annotations_rejected(after)
        else:
            raise ValueError(f"invalid status")

    def _annotations_accepted(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND A.id > (
                SELECT id 
                FROM annotations 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT A.uuid, A.version
            FROM annotations A
            LEFT JOIN annotation_status AST
                ON A.uuid = AST.uuid AND A.version = AST.version 
                    AND AST.status IN (
                        'CREATE_PENDING', 
                        'CREATE_REJECTED', 
                        'DELETE_ACCEPTED')
            WHERE AST.status IS NULL
                { where }
            ORDER BY A.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version, in res]

    def _annotations_pending(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """WHERE A.id > (
                SELECT id 
                FROM annotations 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT DISTINCT A.uuid, A.version
            FROM annotations A
            INNER JOIN annotation_status AST
                ON A.uuid = AST.uuid AND A.version = AST.version 
                    AND AST.status IN (
                        'CREATE_PENDING', 
                        'DELETE_PENDING')
            { where }
            ORDER BY A.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version, in res]

    def _annotations_deleted(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND A.id > (
                SELECT id 
                FROM annotations 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT A.uuid, A.version
            FROM annotations A
            JOIN annotation_status AST 
                ON A.uuid = AST.uuid 
                    AND A.version = AST.version
            WHERE AST.status = 'DELETE_ACCEPTED'
                { where }
            ORDER BY A.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version, in res]

    def _annotations_rejected(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND A.id > (
                SELECT id 
                FROM annotations 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT A.uuid, A.version
            FROM annotations A
            JOIN annotation_status AST 
                ON A.uuid = AST.uuid 
                    AND A.version = AST.version
            WHERE AS.status = 'CREATE_REJECTED'
                { where }
            ORDER BY A.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version, in res]

    def annotation(self, identifier: events.Identifier):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT A.annotation
                FROM annotations A
                WHERE A.uuid = ?
                    AND A.version = ?""", 
            (str(identifier.uuid), identifier.version))
        res = cur.fetchone()
        con.close()

        if res is None:
            return None

        annotation_json, = res
        annotation_data = json.loads(annotation_json)

        return events.Annotation.deserialize(annotation_data)

    def objects_all(self, 
        uuid_: None|uuid.UUID = None, after: None|uuid.UUID = None):

        if after is not None and uuid_ is not None:
            raise ValueError("only provide one of uuid_, after")

        params: tuple = tuple()
        where = "WHERE S.uuid IS NULL"
        if uuid_ is not None:
            where += " AND O.uuid = ?"
            params = (str(uuid_),)

        if after is not None:
            where += """ AND O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT DISTINCT O.uuid, 
                COUNT(O.version) OVER (PARTITION BY O.uuid),
                FIRST_VALUE(O.id) OVER (PARTITION BY O.uuid)
            FROM objects O
            LEFT JOIN schemas S
                ON O.uuid = S.uuid AND O.version = S.version
            {where} 
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [interfaces.ObjectInfo(uuid.UUID(uu), ver) for uu, ver, _ in res]

    def objects_by_annotation(self, annotation_uuid: uuid.UUID):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT OAL.object_uuid, OAL.object_version
                FROM object_annotation_link OAL
                INNER JOIN objects O
                    ON OAL.object_uuid = O.uuid
                        AND OAL.object_version = O.version
                WHERE annotation_uuid = ?""", 
            (str(annotation_uuid),))

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version in res]

    def objects_by_status(self, status: str, after: None|uuid.UUID = None):
        if status == "pending":
            return self._objects_pending(after)
        elif status == "accepted":
            return self._objects_accepted(after)
        elif status == "deleted":
            return self._objects_deleted(after)
        elif status == "rejected":
            return self._objects_rejected(after)
        else:
            raise ValueError(f"invalid status")

    def _objects_accepted(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT O.uuid, O.version
            FROM objects O
            LEFT JOIN schemas S
                ON O.uuid = S.uuid AND O.version = S.version
            LEFT JOIN object_status OS
                ON O.uuid = OS.uuid AND O.version = OS.version 
                    AND OS.status IN (
                        'CREATE_PENDING', 
                        'CREATE_REJECTED', 
                        'DELETE_ACCEPTED')
            WHERE OS.status IS NULL
                AND S.uuid IS NULL
                { where }
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version in res]

    def _objects_pending(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT DISTINCT O.uuid, O.version
            FROM objects O
            LEFT JOIN schemas S
                ON O.uuid = S.uuid AND O.version = S.version
            INNER JOIN object_status OS
                ON O.uuid = OS.uuid AND O.version = OS.version 
                    AND OS.status IN (
                        'CREATE_PENDING', 
                        'DELETE_PENDING')
            WHERE S.uuid IS NULL
                { where }
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version in res]

    def _objects_deleted(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT O.uuid, O.version
            FROM objects O
            INNER JOIN object_status OS 
                ON O.uuid = OS.uuid 
                    AND O.version = OS.version
            WHERE OS.status = 'DELETE_ACCEPTED'
                { where }
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version in res]

    def _objects_rejected(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT O.uuid, O.version
            FROM objects O
            INNER JOIN object_status OS 
                ON O.uuid = OS.uuid 
                    AND O.version = OS.version
            WHERE OS.status = 'CREATE_REJECTED'
                { where }
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [events.Identifier(uuid.UUID(uuid_), version) 
            for uuid_, version in res]

    def object(self, identifier: events.Identifier):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT O.object
                FROM objects O
                LEFT JOIN schemas S
                    ON O.uuid = S.uuid AND O.version = S.version
                WHERE O.uuid = ?
                    AND O.version = ?
                    AND S.uuid IS NULL""", 
            (str(identifier.uuid), identifier.version))
        res = cur.fetchone()
        con.close()

        if res is None:
            return None

        object_json, = res
        object_data = json.loads(object_json)

        return events.Object.deserialize(object_data)

    def schemas_all(self, name: None|str =None): 
        params: tuple = tuple()
        where = ""
        if name is not None:
            where = " WHERE name = ?"
            params = (name,)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT DISTINCT name, uuid, 
                COUNT(version) OVER (PARTITION BY uuid)
            FROM schemas""" + where, params)

        res = cur.fetchall()
        con.close()

        return [interfaces.SchemaInfo(name, uuid.UUID(uuid_), versions) 
            for name, uuid_, versions in res]

    def schemas_by_status(self, status: str, after: None|uuid.UUID = None):
        if status == "pending":
            return self._schemas_pending(after)
        elif status == "accepted":
            return self._schemas_accepted(after)
        elif status == "deprecated":
            return self._schemas_deprecated(after)
        elif status == "rejected":
            return self._schemas_rejected(after)
        else:
            raise ValueError(f"invalid status")

    def _schemas_accepted(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT O.uuid, O.version, O.object->>'$.name'
            FROM objects O
            INNER JOIN schemas S
                ON O.uuid = S.uuid AND O.version = S.version
            LEFT JOIN object_status OS
                ON O.uuid = OS.uuid AND O.version = OS.version 
                    AND OS.status IN (
                        'CREATE_PENDING', 
                        'CREATE_REJECTED', 
                        'DELETE_ACCEPTED')
            WHERE OS.status IS NULL
                { where }
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [interfaces.NamedIdentifier(uuid.UUID(uuid_), version, name) 
            for uuid_, version, name in res]

    def _schemas_pending(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """WHERE O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT DISTINCT O.uuid, O.version, O.object->>'$.name'
            FROM objects O
            INNER JOIN schemas S
                ON O.uuid = S.uuid AND O.version = S.version
            INNER JOIN object_status OS
                ON O.uuid = OS.uuid AND O.version = OS.version 
                    AND OS.status IN (
                        'CREATE_PENDING', 
                        'DELETE_PENDING')
            { where }
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [interfaces.NamedIdentifier(uuid.UUID(uuid_), version, name) 
            for uuid_, version, name in res]

    def _schemas_deprecated(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT O.uuid, O.version, O.object->>'$.name'
            FROM objects O
            INNER JOIN schemas S
                ON O.uuid = S.uuid AND O.version = S.version
            INNER JOIN object_status OS 
                ON O.uuid = OS.uuid 
                    AND O.version = OS.version
            WHERE OS.status = 'DELETE_ACCEPTED'
                { where }
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [interfaces.NamedIdentifier(uuid.UUID(uuid_), version, name) 
            for uuid_, version, name in res]

    def _schemas_rejected(self, after: None|uuid.UUID = None):
        params: tuple = tuple()
        where = ""
        if after is not None:
            where = """AND O.id > (
                SELECT id 
                FROM objects 
                WHERE uuid = ?
                ORDER BY id 
                LIMIT 1)"""
            params += (str(after),)

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(f"""SELECT O.uuid, O.version, O.object->>'$.name'
            FROM objects O
            INNER JOIN schemas S
                ON O.uuid = S.uuid AND O.version = S.version
            INNER JOIN object_status OS 
                ON O.uuid = OS.uuid 
                    AND O.version = OS.version
            WHERE OS.status = 'CREATE_REJECTED'
                { where }
            ORDER BY O.id
            LIMIT 25""", params)

        res = cur.fetchall()
        con.close()

        return [interfaces.NamedIdentifier(uuid.UUID(uuid_), version, name) 
            for uuid_, version, name in res]

    def schema(self, name: str, version: int) -> None|events.Object:
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT O.object
                FROM schemas S
                INNER JOIN objects O
                    ON S.uuid = O.uuid AND S.version = O.version
                WHERE S.name = ?
                    AND S.version = ?""", 
            (name, version))
        res = cur.fetchone()
        con.close()

        if res is None:
            return None

        schema_json, = res
        schema_data = json.loads(schema_json)

        return events.Object.deserialize(schema_data)

    def owners(self): 
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT owner FROM owners ORDER BY id""")
        res = cur.fetchall()
        con.close()

        return [owner for owner, in res]

    def consume(self, event: events.EventT):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute(
            "INSERT INTO events (uuid, type) VALUES (?, ?)",
            (str(event.uuid), event.__class__.__name__))
        con.commit()
        con.close()

        super().consume(event)

    def _consume_object_create(self, event: events.ObjectCreateEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO objects
                (uuid, version, object)
                VALUES (?, ?, ?)""",
            (str(event.object.uuid),
                event.object.version,
                json.dumps(event.object.serialize())))

        if validators.is_schema(event.object.name):
            cur.execute("""INSERT INTO schemas
                    (name, uuid, version)
                    VALUES (?, ?, ?)""",
                (event.object.name,
                    str(event.object.uuid),
                    event.object.version))

        cur.execute("""INSERT INTO object_event_link
                (object_uuid, object_version, event_uuid)
                VALUES (?, ?, ?)""",
            (str(event.object.uuid), event.object.version, str(event.uuid)))

        cur.execute("""INSERT INTO object_status
                (uuid, version, status)
                VALUES (?, ?, ?)""",
            (str(event.object.uuid),
                event.object.version,
                StatusT.CREATE_PENDING.name))

        con.commit()
        con.close()

    def _consume_object_update(self, event: events.ObjectUpdateEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO objects
                (uuid, version, object)
                VALUES (?, ?, ?)""",
            (str(event.object.uuid),
                event.object.version,
                json.dumps(event.object.serialize())))

        if validators.is_schema(event.object.name):
            cur.execute("""INSERT INTO schemas
                    (name, uuid, version)
                    VALUES (?, ?, ?)""",
                (event.object.name,
                    str(event.object.uuid),
                    event.object.version))

        cur.execute("""INSERT INTO object_event_link
                (object_uuid, object_version, event_uuid)
                VALUES (?, ?, ?)""",
            (str(event.object.uuid), event.object.version, str(event.uuid)))

        cur.execute("""INSERT INTO object_status
                (uuid, version, status)
                VALUES (?, ?, ?)""",
            (str(event.object.uuid),
                event.object.version,
                StatusT.CREATE_PENDING.name))

        con.commit()
        con.close()

    def _consume_object_delete(self, event: events.ObjectDeleteEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO object_event_link
                (object_uuid, object_version, event_uuid)
                VALUES (?, ?, ?)""",
            (str(event.object_identifier.uuid),
                event.object_identifier.version,
                str(event.uuid)))

        cur.execute("""INSERT INTO object_status
                (uuid, version, status)
                VALUES (?, ?, ?)""",
            (str(event.object_identifier.uuid),
                event.object_identifier.version,
                StatusT.DELETE_PENDING.name))

        con.commit()
        con.close()

    def _consume_annotation_create(self, event: events.AnnotationCreateEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO annotations
                (uuid, version, annotation)
                VALUES (?, ?, ?)""",
            (str(event.annotation.uuid),
                event.annotation.version,
                json.dumps(event.annotation.serialize())))

        for identifier in event.object_identifiers:
            cur.execute("""INSERT INTO object_annotation_link
                    (object_uuid, object_version, annotation_uuid)
                    VALUES (?, ?, ?)""",
                (str(identifier.uuid),
                    identifier.version,
                    str(event.annotation.uuid)))

        cur.execute("""INSERT INTO annotation_event_link
                (annotation_uuid, annotation_version, event_uuid)
                VALUES (?, ?, ?)""",
            (str(event.annotation.uuid),
                event.annotation.version,
                str(event.uuid)))

        cur.execute("""INSERT INTO annotation_status
                (uuid, version, status)
                VALUES (?, ?, ?)""",
            (str(event.annotation.uuid),
                event.annotation.version,
                StatusT.CREATE_PENDING.name))

        con.commit()
        con.close()

    def _consume_annotation_update(self, event: events.AnnotationUpdateEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO annotations
                (uuid, version, annotation)
                VALUES (?, ?, ?)""",
            (str(event.annotation.uuid),
                event.annotation.version,
                json.dumps(event.annotation.serialize())))

        cur.execute("""INSERT INTO annotation_event_link
            (annotation_uuid, annotation_version, event_uuid)
            VALUES (?, ?, ?)""",
            (str(event.annotation.uuid),
                event.annotation.version,
                str(event.uuid)))

        cur.execute("""INSERT INTO annotation_status
            (uuid, version, status)
            VALUES (?, ?, ?)""",
            (str(event.annotation.uuid),
                event.annotation.version,
                StatusT.CREATE_PENDING.name))

        con.commit()
        con.close()

    def _consume_annotation_delete(self, event: events.AnnotationDeleteEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO annotation_event_link
            (annotation_uuid, annotation_version, event_uuid)
            VALUES (?, ?, ?)""",
            (str(event.annotation_identifier.uuid),
                event.annotation_identifier.version,
                str(event.uuid)))

        cur.execute("""INSERT INTO annotation_status
            (uuid, version, status)
            VALUES (?, ?, ?)""",
            (str(event.annotation_identifier.uuid),
                event.annotation_identifier.version,
                StatusT.DELETE_PENDING.name))

        con.commit()
        con.close()

    def _consume_review_accept(self, event: events.ReviewAcceptEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO event_review_link
                (event_uuid, review_uuid, accepted)
                VALUES (?, ?, ?)""",
            (str(event.event_uuid), str(event.uuid), True))

        target = self.record_keeper.read(event.event_uuid)
        if isinstance(target, events.ObjectEvent):
            if isinstance(target,
                (events.ObjectCreateEvent, events.ObjectUpdateEvent)):

                link_params = (str(target.object.uuid),
                    target.object.version,
                    str(event.uuid))

                cur.execute("""DELETE FROM object_status
                    WHERE uuid = ?
                        AND version = ?
                        AND status = ?""",
                    (str(target.object.uuid),
                        target.object.version,
                        StatusT.CREATE_PENDING.name))
            elif isinstance(target, events.ObjectDeleteEvent):
                link_params = (str(target.object_identifier.uuid),
                    target.object_identifier.version,
                    str(event.uuid))

                cur.execute("""DELETE FROM object_status
                    WHERE uuid = ?
                        AND version = ?
                        AND status = ?""",
                    (str(target.object_identifier.uuid),
                        target.object_identifier.version,
                        StatusT.DELETE_PENDING.name))

                cur.execute("""INSERT INTO object_status
                    (uuid, version, status)
                    VALUES (?, ?, ?)""",
                    (str(target.object_identifier.uuid),
                        target.object_identifier.version,
                        StatusT.DELETE_ACCEPTED.name))
            else:
                raise ValueError("unexpected object event type in accept")

            cur.execute("""INSERT INTO object_event_link
                (object_uuid, object_version, event_uuid)
                VALUES (?, ?, ?)""", link_params)
        elif isinstance(target, events.AnnotationEvent):
            if isinstance(target,
                (events.AnnotationCreateEvent, events.AnnotationUpdateEvent)):

                link_params = (str(target.annotation.uuid),
                    target.annotation.version,
                    str(event.uuid))

                cur.execute("""DELETE FROM annotation_status
                    WHERE uuid = ?
                        AND version = ?
                        AND status = ?""",
                    (str(target.annotation.uuid),
                        target.annotation.version,
                        StatusT.CREATE_PENDING.name))
            elif isinstance(target, events.AnnotationDeleteEvent):
                link_params = (str(target.annotation_identifier.uuid),
                    target.annotation_identifier.version,
                    str(event.uuid))

                cur.execute("""DELETE FROM annotation_status
                    WHERE uuid = ?
                        AND version = ?
                        AND status = ?""",
                    (str(target.annotation_identifier.uuid),
                        target.annotation_identifier.version,
                        StatusT.DELETE_PENDING.name))

                cur.execute("""INSERT INTO annotation_status
                    (uuid, version, status)
                    VALUES (?, ?, ?)""",
                    (str(target.annotation_identifier.uuid),
                        target.annotation_identifier.version,
                        StatusT.DELETE_ACCEPTED.name))
            else:
                raise ValueError("unexpected annotation event type in accept")

            cur.execute("""INSERT INTO annotation_event_link
                (annotation_uuid, annotation_version, event_uuid)
                VALUES (?, ?, ?)""", link_params)
        else:
            raise ValueError("unexpected event type in accept")
        
        con.commit()
        con.close()

    def _consume_review_reject(self, event: events.ReviewRejectEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO event_review_link
            (event_uuid, review_uuid, accepted)
            VALUES (?, ?, ?)""",
            (str(event.event_uuid), str(event.uuid), False))

        target = self.record_keeper.read(event.event_uuid)
        if isinstance(target, events.ObjectEvent):
            if isinstance(target,
                (events.ObjectCreateEvent, events.ObjectUpdateEvent)):

                link_params = (str(target.object.uuid),
                    target.object.version,
                    str(event.uuid))

                cur.execute("""DELETE FROM object_status
                    WHERE uuid = ?
                        AND version = ?
                        AND status = ?""",
                    (str(target.object.uuid),
                        target.object.version,
                        StatusT.CREATE_PENDING.name))

                cur.execute("""INSERT INTO object_status
                    (uuid, version, status)
                    VALUES (?, ?, ?)""",
                    (str(target.object.uuid),
                        target.object.version,
                        StatusT.CREATE_REJECTED.name))
            elif isinstance(target, events.ObjectDeleteEvent):
                link_params = (str(target.object_identifier.uuid),
                    target.object_identifier.version,
                    str(event.uuid))

                cur.execute("""DELETE FROM object_status
                    WHERE uuid = ?
                        AND version = ?
                        AND status = ?""",
                    (str(target.object_identifier.uuid),
                        target.object_identifier.version,
                        StatusT.DELETE_PENDING.name))
            else:
                raise ValueError("unexpected object event type in reject")

            cur.execute("""INSERT INTO object_event_link
                (object_uuid, object_version, event_uuid)
                VALUES (?, ?, ?)""", link_params)
        elif isinstance(target, events.AnnotationEvent):
            if isinstance(target, (
                events.AnnotationCreateEvent, events.AnnotationUpdateEvent)):

                link_params = (str(target.annotation.uuid),
                    target.annotation.version,
                    str(event.uuid))

                cur.execute("""DELETE FROM annotation_status
                    WHERE uuid = ?
                        AND version = ?
                        AND status = ?""",
                    (str(target.annotation.uuid),
                        target.annotation.version,
                        StatusT.CREATE_PENDING.name))

                cur.execute("""INSERT INTO annotation_status
                    (uuid, version, status)
                    VALUES (?, ?, ?)""",
                    (str(target.annotation.uuid),
                        target.annotation.version,
                        StatusT.CREATE_REJECTED.name))
            elif isinstance(target, events.AnnotationDeleteEvent):
                link_params = (str(target.annotation_identifier.uuid),
                    target.annotation_identifier.version,
                    str(event.uuid))

                cur.execute("""DELETE FROM annotation_status
                    WHERE uuid = ?
                        AND version = ?
                        AND status = ?""",
                    (str(target.annotation_identifier.uuid),
                        target.annotation_identifier.version,
                        StatusT.DELETE_PENDING.name))
            else:
                raise ValueError("unexpected annotation event type in reject")

            cur.execute("""INSERT INTO annotation_event_link
                (annotation_uuid, annotation_version, event_uuid)
                VALUES (?, ?, ?)""", link_params)
        else:
            raise ValueError("unexpected event type in reject")

        con.commit()
        con.close()

    def _consume_owner_add(self, event: events.OwnerAddEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""INSERT INTO owners (owner) VALUES (?)""",
            (event.owner,))

        con.commit()
        con.close()

    def _consume_owner_remove(self, event: events.OwnerRemoveEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""DELETE FROM owners WHERE owner = ?""",
            (event.owner,))

        con.commit()
        con.close()

    def _validate_object_create(self, event: events.ObjectCreateEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        if validators.is_schema(event.object.name):
            cur.execute("""SELECT COUNT(*) FROM schemas WHERE name = ?""",
                (event.object.name,))
            count, = cur.fetchone()
            if count != 0:
                con.close()
                raise exceptions.ValidationError("schema name already in use")

        cur.execute("""SELECT COUNT(*)
                FROM objects
                WHERE uuid = ?""",
            (str(event.object.uuid),))

        count, = cur.fetchone()
        if count != 0:
            con.close()
            raise exceptions.ValidationError("object with UUID already exists")

        cur.execute("""SELECT uuid, version
                FROM objects
                WHERE object->>'$.hash' = ?
                    AND uuid != ?""",
            (event.object.hash, str(event.object.uuid)))

        res = cur.fetchone()
        con.close()

        if res is not None:
            dup_uuid, dup_version = res
            raise exceptions.ValidationError(
                f"duplicate hash detected in object {dup_uuid}:{dup_version}")

        if event.object.version != 0:
            raise exceptions.ValidationError(
                "object version must be zero in create event")

    def _validate_object_update(self, event: events.ObjectCreateEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT uuid, version
                FROM objects
                WHERE object->>'$.hash' = ?
                    AND uuid != ?""",
            (event.object.hash, str(event.object.uuid)))

        res = cur.fetchone()

        if res is not None:
            con.close()
            dup_uuid, dup_version = res
            raise exceptions.ValidationError(
                f"duplicate hash detected in object {dup_uuid}:{dup_version}")

        cur.execute("""SELECT object
                FROM objects
                WHERE uuid = ?""",
            (str(event.object.uuid),))

        versions_json = cur.fetchall()
        con.close()
        if len(versions_json) == 0:
            raise exceptions.ValidationError("no objects with UUID found")

        versions_data = [json.loads(ea) for ea, in versions_json]

        versions = [events.Object.deserialize(ea) for ea in versions_data]

        if validators.is_schema(versions[-1].name):
            if versions[-1].name != event.object.name:
                raise exceptions.ValidationError("schema names may not change")
        else:
            if validators.is_schema(event.object.name):
                raise exceptions.ValidationError("object may not become schema")

        if versions[-1].hash == event.object.hash:
            raise exceptions.ValidationError("object hash unchanged")

        if event.object.version != len(versions):
            raise exceptions.ValidationError(
                f"object version should be {len(versions)}")

    def _validate_object_delete(self, event: events.ObjectDeleteEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT COUNT(*)
                FROM objects
                WHERE uuid = ?
                    AND version = ?""",
            (str(event.object_identifier.uuid),
                event.object_identifier.version))

        count, = cur.fetchone()
        if count == 0:
            con.close()
            raise exceptions.ValidationError("object identifier not found")

        cur.execute("""SELECT status
                FROM object_status
                WHERE uuid = ?
                    AND version = ?""",
            (str(event.object_identifier.uuid),
                event.object_identifier.version))

        status = {getattr(StatusT, ea) for ea, in cur.fetchall()}
        con.close()
        if StatusT.CREATE_REJECTED in status:
            raise exceptions.ValidationError("cannot delete a rejected object")

        if StatusT.DELETE_PENDING in status:
            raise exceptions.ValidationError("object version pending deletion")

        if StatusT.DELETE_ACCEPTED in status:
            raise exceptions.ValidationError("object version already deleted")

    def _validate_annotation_create(self, event: events.AnnotationCreateEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT COUNT(*)
                FROM annotations
                WHERE uuid = ?""",
            (str(event.annotation.uuid),))

        count, = cur.fetchone()
        if count != 0:
            con.close()
            raise exceptions.ValidationError(
                "annotation with UUID already exists")

        if event.annotation.version != 0:
            con.close()
            raise exceptions.ValidationError(
                "annotation version must be zero in create event")

        for identifier in event.object_identifiers:
            cur.execute("""SELECT COUNT(*)
                    FROM objects
                    WHERE uuid = ?
                        AND version = ?""",
                (str(identifier.uuid), identifier.version))
            count, = cur.fetchone()
            if count == 0:
                con.close()
                raise exceptions.ValidationError("object identifier not found")

            cur.execute("""SELECT status
                FROM object_status
                WHERE uuid = ?
                    AND version = ?""",
                (str(identifier.uuid), identifier.version))

            status = {getattr(StatusT, ea) for ea, in cur.fetchall()}
            if StatusT.CREATE_REJECTED in status:
                con.close()
                raise exceptions.ValidationError(
                    "rejected objects cannot be annotated")

            if StatusT.DELETE_ACCEPTED in status:
                con.close()
                raise exceptions.ValidationError(
                    "deleted objects cannot be annotated")

            cur.execute("""SELECT COUNT(*) FROM schemas WHERE uuid = ?""",
                (str(identifier.uuid),))
            count, = cur.fetchone()
            con.close()
            if count != 0:
                raise exceptions.ValidationError("schemas can not be annotated")

    def _validate_annotation_update(self, event: events.AnnotationUpdateEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT uuid, version
                FROM annotations
                WHERE uuid = ?""",
            (str(event.annotation.uuid),))

        version_ids = cur.fetchall()
        con.close()
        if len(version_ids) == 0:
            raise exceptions.ValidationError("no annotations with UUID found")

        if event.annotation.version != len(version_ids):
            raise exceptions.ValidationError(
                f"annotation version should be {len(version_ids)}.")

    def _validate_annotation_delete(self, event: events.AnnotationDeleteEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT COUNT(*)
                FROM annotations
                WHERE uuid = ?
                    AND version = ?""",
            (str(event.annotation_identifier.uuid),
                event.annotation_identifier.version))

        count, = cur.fetchone()
        if count == 0:
            con.close()
            raise exceptions.ValidationError("annotation identifier not found")

        cur.execute("""SELECT status
                FROM annotation_status
                WHERE uuid = ?
                    AND version = ?""",
            (str(event.annotation_identifier.uuid),
                event.annotation_identifier.version))

        status = {getattr(StatusT, ea) for ea, in cur.fetchall()}
        con.close()
        if StatusT.CREATE_REJECTED in status:
            raise exceptions.ValidationError(
                "cannot delete a rejected annotation")

        if StatusT.DELETE_PENDING in status:
            raise exceptions.ValidationError(
                "annotation already pending deletion")

        if StatusT.DELETE_ACCEPTED in status:
            raise exceptions.ValidationError("annotation already deleted")

    def _validate_review(self,
        event: events.ReviewAcceptEvent|events.ReviewRejectEvent):

        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT COUNT(*)
                FROM event_review_link
                WHERE event_uuid = ?""",
            (str(event.event_uuid),))

        count, = cur.fetchone()
        if count != 0:
            con.close()            
            raise exceptions.ValidationError("event already reviewed")

        if not self.record_keeper.exists(event.event_uuid):
            con.close()            
            raise exceptions.ValidationError("no events with event UUID found")

        target_event = self.record_keeper.read(event.event_uuid)
        if not isinstance(target_event,
            (events.AnnotationEvent, events.ObjectEvent)):
            con.close()        
            raise exceptions.ValidationError(
                "review on non object or annotation event")

        if event.author is None:
            con.close()            
            raise exceptions.ValidationError("author is empty")

        cur.execute("""SELECT COUNT(*)
            FROM owners
            WHERE owner = ?""",
            (event.author,))
        count, = cur.fetchone()
        con.close()
        if count == 0:
            raise exceptions.ValidationError("review event from non-owner")

    def _validate_review_accept(self, event: events.ReviewAcceptEvent):
        self._validate_review(event)

    def _validate_review_reject(self, event: events.ReviewRejectEvent):
        self._validate_review(event)

    def _validate_owner_add(self, event: events.OwnerAddEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT owner FROM owners""")

        owners = {owner for owner, in cur.fetchall()}
        con.close()
        if len(owners) > 0:
            if event.owner in owners:
                raise exceptions.ValidationError("owner already present")

            if event.author not in owners:
                raise exceptions.ValidationError("only owners can add owners")
        else:
            if event.owner != event.author:
                raise exceptions.ValidationError(
                    "first owner add event must be self signed")

    def _validate_owner_remove(self, event: events.OwnerRemoveEvent):
        con = sqlite3.connect(self.database_path)
        cur = con.cursor()
        cur.execute("""SELECT id, owner FROM owners""")

        owners = cur.fetchall()
        con.close()
        if len(owners) == 0:
            raise exceptions.ValidationError("dataset has no owners to remove")

        ranks = {owner: id_ for id_, owner in owners}

        if event.author not in ranks:
            raise exceptions.ValidationError("only owners may remove owners")

        if event.owner not in ranks:
            raise exceptions.ValidationError("target key is not an owner")

        if len(ranks) == 1:
            raise exceptions.ValidationError(
                "removing owner would leave the dataset ownerless")

        target_rank = ranks[event.owner]
        signer_rank = ranks[event.author]

        if signer_rank > target_rank:
            raise exceptions.ValidationError(
                "cannot remove a higher ranking owner")



# Copyright 2023 - Compute Heavy Industries Incorporated
# This work is released, distributed, and licensed under AGPLv3.

from setuptools import setup

setup(
    name='gonk',
    version='0.1',
    packages=['gonk', 'gonk.core', 'gonk.api', 'gonk.impl'],
    package_dir={
        'gonk': 'src',
        'gonk.core': 'src/core',
        'gonk.api': 'src/api',
        'gonk.impl': 'src/impl',
    },
    entry_points={
        'console_scripts': [
            'gonk-api=gonk.api.server:cli',
        ],
    },
)